/* vrJuggler interface
 * Copyright (C) 2001 Paul Rajlich
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef WIN32
#include <windows.h>
#endif
#include <GL/gl.h>


// vrJuggler stuff
#include <vrj/Kernel/Kernel.h>
#include <vrj/Draw/OGL/GlApp.h>
#include <vrj/Draw/OGL/GlContextData.h>
#include <vrj/Util/Debug.h>
#include <vrj/Display/Projection.h>
#include <gadget/Type/PositionInterface.h>
#include <gadget/Type/DigitalInterface.h>
#include <gadget/Type/AnalogInterface.h>

#include <gmtl/Vec.h>
#include <gmtl/Generate.h>
#include <gmtl/Output.h>

#include "vec.h"
#include "uicommon.h"
#include "rendercontext.h"
#include "globalshared.h"

global_shared_t *g;

void printUsage();

// use regular memory (juggler uses threads)
void *gc_malloc(size_t size) { return malloc(size); }
void *rc_malloc(size_t size) { return malloc(size); }
void gc_free(void *mem) { free(mem); }
void rc_free(void *mem) { free(mem); }

// timing
#ifdef WIN32
double gettime()
{
   return ((double)GetTickCount())/1000.0f;
}
#else
double gettime()
{
    struct timeval tv;
    gettimeofday(&tv, 0);
    return  ((double) tv.tv_sec + (double) tv.tv_usec / 1000000.0);
}
#endif


// Wrapper class to hold a copy of a context specific data structure
class ContextData
{
public:
   ContextData() : r_context(NULL)
   {
      r_context = (r_context_t *) rc_malloc(sizeof(r_context_t));
      std::cout << "Allocated ContextData." << std::endl;
   }

   r_context_t* getData()
   { return r_context; }

private:
   r_context_t* r_context;
};


// Actual application class
// Just derives from vjGlApp base class and implments the interface methods
// needed to call the rendering engine.
class myApp : public vrj::GlApp {
private:
  // see rendercontext.h. Each rendering thread will have its own copy
  vrj::GlContextData<ContextData> contextData;

public:
  gadget::PositionInterface mWand;
  gadget::PositionInterface mHead;
  gadget::AnalogInterface mAnalog0;
  gadget::AnalogInterface mAnalog1;
  gadget::DigitalInterface mButton0;
  gadget::DigitalInterface mButton1;
  gadget::DigitalInterface mButton2;
  gadget::DigitalInterface mButton3;
  gadget::DigitalInterface mButton4;

  double starttime, frametime, lastTime;

public:
  myApp(vrj::Kernel *kern) : vrj::GlApp(kern) {
    mWand.init("VJWand");
    mHead.init("VJHead");
    mAnalog0.init("VJAnalog0");
    mAnalog1.init("VJAnalog1");
    mButton0.init("VJButton0");
    mButton1.init("VJButton1");
    mButton2.init("VJButton2");
    mButton3.init("VJButton3");
    mButton4.init("VJButton4");
    //std::cerr << std::endl;
    //std::cerr << "***************************************" << std::endl;
    //std::cerr << "bsp: " << bsp << " pak: " << pak << std::endl;
    //std::cerr << "***************************************" << std::endl;
    //std::cerr << std::endl;
  }

   // Initialize global data structures
  virtual void init() {

    // load the map
    ui_init_bsp();
    printf("\n\n");

    starttime = gettime();
    frametime = starttime;
    lastTime = starttime;

    // juggler will handle projection, origin is not at eye but at ground!
    g->r_setup_projection = 0;
    g->r_origin_offset = -(g->g_ground_offset);
  }

  // Initialize context data
  virtual void contextInit() {
    r_context_t* the_context_data = contextData->getData();
    vprASSERT(the_context_data != NULL && "context data pointer should be non-NULL");

    ui_init_gl(the_context_data);
  }

  // Execute the actually drawing
  virtual void draw() {
    r_context_t* the_context_data = contextData->getData();
    vprASSERT(the_context_data != NULL);

    // Set some state that is expected
    glDisable(GL_LIGHTING);
    glDisable(GL_LIGHT0);
    glCullFace(GL_FRONT);
    glShadeModel(GL_SMOOTH);

    ui_display(the_context_data, frametime - starttime);
    glDisable(GL_BLEND);

    // reset some state to make simulator mode look "better"
    glCullFace(GL_BACK);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glShadeModel(GL_FLAT);
  }

  // Updates that occur immedately after device updates but before rendering
  // - Update timing
  // - Update navigation
  virtual void preFrame() {
    double etime;
    float wandH(0.0), wandP(0.0), wandR(0.0);

    frametime = gettime();
    etime = frametime - lastTime;
    if (etime > 0.2f)
      etime = 0.2f; /* cap */

    // move forward
    if (mButton0->getData())
      ui_key_down('w');
    else
      ui_key_up('w');

    // change sky
    if (mButton4->getData() == gadget::Digital::TOGGLE_ON)
      ui_key_down('b');

    // turn on/off collision detection
    if (mButton3->getData() == gadget::Digital::TOGGLE_ON)
      ui_key_down('m');

    if (mButton1->getData() == gadget::Digital::TOGGLE_ON)
      ui_key_down('f');

    if (mButton2->getData() == gadget::Digital::TOGGLE_ON)
      ui_key_down('x');

    // reset
    //if (mButton0->getData() && mButton1->getData())
    //  ui_reset();

#if 0 // head offset for falling, see uicommon.c (NOT WORKING!!)
    gmtl::Matrix44f headMat = mHead->getData();
	gmtl::Matrix44f rotated_head;   // head position with correction for current rotation
	gmtl::Matrix44f correction_mat; // matrix to account for world rotation (g->r_eye_az)
	gmtl::setRot(correction_mat, gmtl::EulerAngleXYZf(0.0f, g->r_eye_az*DEG2RAD, 0.0f));
	gmtl::mult(rotated_head, correction_mat, headMat);
	gmtl::Vec3f trans;
	gmtl::setTrans(trans, rotated_head);
	g->g_walk_headOffsetX = trans[0];
	g->g_walk_headOffsetY = -trans[2]; // y in quake is -z in juggler
	//std::cerr << "az " << g->r_eye_az << " headOffset " << g->g_walk_headOffsetX;
	//std::cerr << " " << g->g_walk_headOffsetY << std::endl;
#endif
	
    // travel in direction wand is pointing
    gmtl::Matrix44f wandMat = mWand->getData();
    wandH = (gmtl::makeRot<gmtl::EulerAngleXYZf>(wandMat))[1];
    wandP = (gmtl::makeRot<gmtl::EulerAngleXYZf>(wandMat))[0];
    wandH = wandH/DEG2RAD; // from radians to degrees
    wandP = wandP/DEG2RAD;
    //std::cerr << "wandH: " << wandH << " wandP: " << wandP << std::endl;

    if (g->g_walk) wandP = 0.0f; // only use heading when walking

    vec3_t dir;
    vec_point(dir, wandH + g->r_eye_az, wandP);
    g->r_movedir[0] = dir[0];
    g->r_movedir[1] = dir[1];
    g->r_movedir[2] = dir[2];

    // in juggler analog values are [0.0, 1.0], make it [-1.0, 1.0]
    float joyX = mAnalog0->getData()*2.0f - 1.0f;
    float joyY = mAnalog1->getData()*2.0f - 1.0f;
    if (fabsf(joyX) > 0.3f)
      g->r_eye_az = g->r_eye_az - joyX*etime*50.0f;

    if (joyY > 0.5f)  ui_key_down('w');
    if (joyY < -0.5f) ui_key_down('s');

    ui_move(frametime - starttime);

    lastTime = frametime;
  }
};


int
main(int argc, char **argv)
{
  // setup global data structure
  g = (global_shared_t *) gc_malloc(sizeof(global_shared_t));
  init_global_shared(g);

  // USAGE:
  // command -c configfile1 -c configfile2 -- otherOptions
  // NOTE: The config files all have to be first then the "--" specification
  if(argc <= 1)
    printUsage();

  vrj::Kernel* kernel = vrj::Kernel::instance();

 /** FIX LATER
  // Get the Juggler commands first
  int c;
  while ( (c = getopt(argc, argv, "hc:")) != -1)
  {
      switch (c) {
      case 'c':
           kernel->loadConfigFile( std::string(optarg) );
           break;
      case 'h':
            printUsage();
            exit(1);
            break;
      case '?':
           break;
      }
  }

  // optind-1 because it will ignore elt 0 (normally app name)
  ui_read_args((argc-optind+1), &(argv[optind-1]));
  ***/

  // quick hack: assume one config file only...
  kernel->loadConfigFile(argv[1]);
  
  ui_read_args(argc-1, &argv[1]);
  strcpy(g->r_help_fname, "paul/helpVR.jpg");

  myApp *app = new myApp(kernel);

  kernel->start();
  kernel->setApplication(app);

  kernel->waitForKernelStop();

}

void printUsage()
{
   std::cout << "Usage: cq3a -c <juggler-config-file> -c <more-juggler-config> -- <cq3a config params>" << std::endl;
}

